Require Import LibTactics.
Require Coq.omega.Omega.

Ltac myForwardOmega f :=
  assert f as evidence; [ omega | rewrite evidence; clear evidence; auto ].

Inductive ty : Type :=
| TyLeaf : ty
| TyArrow : ty -> ty -> ty
| TyBullet : ty -> ty.

(* Auxiliary functions over types. *)

Fixpoint size (t : ty) : nat :=
  match t with
  | TyLeaf => 0
  | TyArrow t u => size t + size u
  | TyBullet t => size t + 1
  end.

Fixpoint bullet (i : nat) (t : ty) { struct i } :=
  match i with
  | 0   => t
  | S i => TyBullet (bullet i t)
  end.

Fixpoint push (i : nat) (t : ty) { struct t } :=
  match t with
  | TyLeaf => bullet i TyLeaf
  | TyArrow t u => TyArrow (push i t) (push i u)
  | TyBullet t => push (i + 1) t
  end.

(* Definition sizes (t, u) := size t + size u. *)

Inductive equal : ty -> ty -> Prop :=
| EqLeaf : equal TyLeaf TyLeaf
| EqArrow : forall u1 v1 u2 v2, equal u2 u1 -> equal v1 v2 -> equal (TyArrow u1 v1) (TyArrow u2 v2)
| EqBullet : forall u v, equal u v -> equal (TyBullet u) (TyBullet v)
| EqArrowBulletDown: forall u v, equal (TyBullet (TyArrow u v)) (TyArrow (TyBullet u) (TyBullet v))
| EqArrowBulletUp : forall u v, equal (TyArrow (TyBullet u) (TyBullet v)) (TyBullet (TyArrow u v))
| EqTrans : forall t u v, equal t u -> equal u v -> equal t v.

Hint Constructors equal.

Inductive sub : ty -> ty -> Prop :=
| SubLeaf : sub TyLeaf TyLeaf
| SubArrow : forall u1 v1 u2 v2, sub u2 u1 -> sub v1 v2 -> sub (TyArrow u1 v1) (TyArrow u2 v2)
| SubBullet : forall u v, sub u v -> sub (TyBullet u) (TyBullet v)
| SubBulletIntro : forall t, sub t (TyBullet t)
| SubArrowBulletDown : forall u v, sub (TyBullet (TyArrow u v)) (TyArrow (TyBullet u) (TyBullet v))
| SubArrowBulletUp : forall u v, sub (TyArrow (TyBullet u) (TyBullet v)) (TyBullet (TyArrow u v))
| SubTrans : forall t u v, sub t u -> sub u v -> sub t v.

Hint Constructors sub.

Inductive dsub : ty -> ty -> Prop :=
| DSubLeaf : forall i j, i <= j -> dsub (bullet i TyLeaf) (bullet j TyLeaf)
| DSubArrow : forall k t1 t2 u1 u2, dsub u1 (push k t1) -> dsub (push k t2) u2 -> dsub (TyArrow t1 t2) (TyArrow u1 u2).

Hint Constructors dsub.

(* Subtyping is reflexive. *)

Lemma equal_reflexive: forall t, equal t t.
Proof.
  induction t; eauto.
Qed.

Hint Resolve equal_reflexive.

Lemma equal_symmetric: forall t u, equal t u -> equal u t. 
Proof.
  intros t u e; induction e; eauto.
Qed.

Hint Immediate equal_symmetric.

(*
Require Import Setoid.

Add Relation ty equal
  reflexivity proved by equal_reflexive
  symmetry proved by equal_symmetric
  transitivity proved by EqTrans
as equal_setoid.

Add Morphism TyArrow with signature equal ==> equal ==> equal as tyarrow_morphism.
Proof.
  eauto.
Qed.

Add Morphism TyBullet with signature equal ==> equal as tybullet_morphism.
Proof.
  eauto.
Qed.*)

Lemma sub_reflexive: forall t, sub t t.
Proof.
  induction t; eauto.
Qed.

Hint Resolve sub_reflexive.

Lemma equal_sub: forall t u, equal t u -> sub t u. 
Proof.
  intros t u e; induction e; eauto.
Qed.

Lemma eq_sub_eq:
  forall t u v w,
  equal t u ->
  sub u v ->
  equal v w ->
  sub t w.
Proof.
  eauto using equal_sub.
Qed. 

(*
Add Morphism sub with signature equal ==> equal ==> flip impl as sub_morphism.
Proof.
  eauto using eq_sub_eq.
Qed.
*)

Lemma sub_bullet_intro: forall i j t, i <= j -> sub (bullet i t) (bullet j t).
Proof.
  intros i j t less; induction less; eauto.
Qed.

(* Pushing respects equality. *)

Lemma bullet_right: forall i t, equal (bullet (i + 1) t) (bullet i (TyBullet t)).
Proof.
  induction i; intros; auto.
  simpl. eauto.
Qed.

Lemma bullet_sub: forall i t, sub t (bullet i t).
Proof.
  induction i; eauto.
Qed.

Lemma bullet_arrow: forall i t u, equal (bullet i (TyArrow t u)) (TyArrow (bullet i t) (bullet i u)).
Proof.
  induction i; intros; simpl; eauto.
Qed.

Lemma push_bullet: forall t i, equal (bullet i t) (push i t).
Proof.
  induction t; intro; simpl.
  eauto.
  eauto using bullet_arrow.
  simpl.
  eauto using bullet_right.
Qed.

Lemma equal_push0: forall t, equal t (push 0 t).
Proof.
  intro.
  assert (equal (bullet 0 t) (push 0 t)).
    apply push_bullet.
  eauto.
Qed.

Lemma push_sub: forall i t, sub t (push i t).
Proof.
  intros.
Admitted.
(*  rewrite <- push_bullet.
  apply bullet_sub.
Qed.*)

(* Inversion lemmas. *)

Ltac myGeneralize t :=
  generalize (refl_equal t); (* introduce an equality t = t *)
  pattern t at -1; (* beta-abstract all occurrences, except the first one; we end up with exactly two occurrences in t *)
  generalize t at 2. (* abstract away the last occurrence *)

Ltac myCase t :=
  generalize (refl_equal t);
  pattern t at -1;
  case t.

Lemma equal_push:
  forall t u, equal t u -> forall i, push i t = push i u.
Proof.
  induction 1; simpl; eauto; congruence.
Qed.

Lemma push_equal:
  forall t u, push 0 t = push 0 u -> equal t u.
Proof.
  intros t u e.
  assert (equal (bullet 0 t) (bullet 0 u)).
    eapply EqTrans. eapply push_bullet.
    eapply EqTrans. rewrite e. eapply equal_reflexive.
    eapply equal_symmetric. eapply push_bullet.
  intuition eauto.
Qed.

(* dsub. *)

Lemma dsub_sub:
  forall t u, dsub t u -> sub t u.
Proof.
  intros t u d; induction d.
  apply sub_bullet_intro; auto.
  eapply SubTrans. eapply push_sub.
  simpl. eauto. 
Qed.

Lemma dsub_push_push_sub:
  forall t u, dsub (push 0 t) (push 0 u) -> sub t u.
Proof.
  intros.
  assert (equal t (push 0 t)). apply equal_push0.
  assert (equal (push 0 u) u). eauto using equal_push0.
  assert (sub (push 0 t) (push 0 u)). eauto using dsub_sub.
  eauto using eq_sub_eq.
Qed.

Lemma size_bullet:
  forall j t, size (bullet j t) = j + size t.
Proof.
  induction j; intro t; simpl.
  tauto.
  rewrite IHj. omega.
Qed.

Lemma bullet_injective:
  forall i j t, bullet i t = bullet j t -> i = j.
Proof.
  intros i j t hyp.
  assert (size (bullet i t) = size (bullet j t)) as obvious. congruence.
  generalize obvious. repeat (rewrite size_bullet). intro.
  omega.
Qed.

Lemma bullet_leaf_arrow:
  forall i u v, ~(bullet i TyLeaf = TyArrow u v).
Proof.
  intros.
  myCase i; simpl; congruence.
Qed.

Lemma bullet_bullet:
  forall i j t, bullet i (bullet j t) = bullet (i + j) t.
Proof.
  induction i; simpl; eauto.
  congruence.
Qed.

Lemma push_bullet_is_bullet:
  forall j i, push i (bullet j TyLeaf) = bullet (i + j) TyLeaf.
Proof.
  induction j; intro.
  myForwardOmega (i + 0 = i).
  myForwardOmega (i + S j = S (i + j)). simpl. rewrite IHj.
  myForwardOmega (i + 1 + j = S (i + j)).
Qed.

Lemma push_bullet_is_push:
  forall j i u v, push i (bullet j (TyArrow u v)) = TyArrow (push (i + j) u) (push (i + j) v).
Proof.
  induction j; intros.
  myForwardOmega (i + 0 = i).
  myForwardOmega (i + S j = S (i + j)). simpl. rewrite IHj.
  myForwardOmega (i + 1 + j = S (i + j)).
Qed.
  
Lemma push_push:
  forall t i j, push i (push j t) = push (i + j) t.
Proof.
  induction t; intros i j; simpl.
  apply push_bullet_is_bullet.
  rewrite IHt1. rewrite IHt2. eauto.
  myForwardOmega (i + j + 1 = i + (j + 1)).
Qed.

Lemma push0_push:
  forall t j, push 0 (push j t) = push j t.
Proof.
  intros.
  apply push_push.
Qed.

Lemma dsub_push_push:
  forall t i j, i <= j -> dsub (push i t) (push j t).
Proof.
  induction t; intros i j ij; simpl; eauto.
  apply (DSubArrow (j - i)); rewrite push_push. apply IHt1. omega. apply IHt2. omega.
  apply IHt. omega.
Qed.

Lemma dsub_push:
  forall t u, dsub t u -> forall i, dsub (push i t) (push i u).
Proof.
  intros t u d; induction d; simpl; eauto.
  intro k. repeat (rewrite push_bullet_is_bullet). econstructor. omega.
  intro i. apply (DSubArrow k).
  generalize (IHd1 i). repeat (rewrite push_push). myForwardOmega (i + k = k + i).
  generalize (IHd2 i). repeat (rewrite push_push). myForwardOmega (i + k = k + i).
Qed.

Lemma dsub_push_plus:
  forall j k t u, dsub (push j t) (push k u) -> forall i, dsub (push (i + j) t) (push (i + k) u).
Proof.
  intros.
  repeat (rewrite <- push_push).
  eauto using dsub_push.
Qed.

Definition transitivity_on u :=
  forall i t v, dsub t (push i u) -> dsub (push i u) v -> dsub t v.

Hint Unfold transitivity_on.

Ltac myBulletInjective H :=
  generalize (bullet_injective _ _ _ H); clear H; intro.

Ltac myImpossible :=
  false; eapply bullet_leaf_arrow; eauto.

Ltac myInversions H :=
  inversions H; try solve [ myImpossible ].


Lemma dsub_transitive:
  forall u, transitivity_on u.
Proof.
  induction u; intros i t v tu uv.
  (* Leaf Leaf *)
  myInversions tu; myInversions uv.
  myBulletInjective H.
  myBulletInjective H0.
  econstructor; omega.
  (* Arrow Arrow *)
  myInversions tu; myInversions uv.
  eapply (DSubArrow (k0 + k));
  rewrite push_push in *;
  eauto using dsub_push_plus.
  (* Leaf Leaf *)
  simpl in *.
  eauto.
Qed.

Lemma dsub_push_reflexive:
  forall j t u, push j t = push j u -> dsub (push j t) (push j u).
Proof.
  intros j t u h.
  rewrite h.
  eauto using dsub_push_push.
Qed.

Lemma sub_dsub:
  forall t u, sub t u -> forall j, dsub (push j t) (push j u).
Proof.
  intros t u s; induction s; intro j.
  (* dsub TyLeaf TyLeaf *)
  econstructor; omega.
  (* dsub TyArrow TyArrow *)
  econstructor; rewrite push0_push; eauto.
  (* dsub TyBullet TyBullet *)
  simpl. eauto.
  (* dsub j (j+1) *)
  simpl. apply dsub_push_push. omega.
  (* Bullet Arrow versus Arrow Bullet *)
  eauto using dsub_push_reflexive.
  (* Arrow Bullet versus Bullet Arrow *)
  eauto using dsub_push_reflexive.
  (* Trans *)
  eapply dsub_transitive; eauto.
Qed.

Lemma leaf_leaf_inversion:
  forall i, bullet i TyLeaf = TyLeaf -> i = 0.
Proof.
  intro.
  myCase i.
  trivial.
  simpl.
  congruence.
Qed.

Lemma push_some:
  forall t i, ~(push (1 + i) t = TyLeaf).
Proof.
  induction t; intros; simpl in *.
  congruence.
  congruence.
  eauto.
Qed.

Lemma push_leaf:
  forall u i, push i u = TyLeaf -> i = 0 /\ u = TyLeaf.
Proof.
  intro.
  myCase u; intros; simpl in *.
  eauto using leaf_leaf_inversion.
  congruence.
  false.
  assert (i + 1 = 1 + i) as obvious. omega. rewrite obvious in H0.
  eapply push_some. eauto.
Qed.

Lemma push_arrow_arrow:
  forall i u1 u2 v1 v2,
  TyArrow v1 v2 = push i (TyArrow u1 u2) ->
  v1 = push i u1 /\ v2 = push i u2.
Proof.
  induction i; simpl; intros; split; congruence.
Qed.

Lemma push_arrow:
  forall u i v1 v2,
  TyArrow v1 v2 = push i u ->
  exist j u1 u2,
  u = bullet j (TyArrow u1 u2) /\ v1 = push (i + j) u1 /\ v2 = push (i + j) u2.
Proof.
  induction u; intros.
  (* Leaf *)
  myImpossible.
  (* Arrow *)
  exists 0. exists u1. exists u2.
  generalize (push_arrow_arrow _ _ _ _ _ H); intros [ ? ? ]; subst.
  intuition eauto;
  myForwardOmega (i + 0 = i).
  (* Bullet *) 
  simpl in H.
  generalize (IHu _ _ _ H). intros [ j [ u1 [ u2 [ ? [ ? ? ]]]]].
  exists (1 + j).
  exists u1.
  exists u2.
  subst.
  intuition eauto; myForwardOmega (i + 1 + j = i + (1 + j)).
Qed.

Ltac myPushArrow H :=
  generalize (push_arrow _ _ _ _ H); clear H; intros [ ? [ ? [ ? [ ? [ ? ? ]]]]]; subst.

(* (* TEMPORARY *)
Lemma leaf_bullet_inversion:
  forall i u, bullet i TyLeaf = TyBullet u ->
  exists j, i = j + 1 /\ bullet j TyLeaf = u.
Proof.
  intro i.
  myCase i.
  simpl. congruence.
  intros n _ u e.
  exists n.
  split. omega.
  simpl in e.
  congruence.
Qed.

Lemma leaf_inversion_left:
  forall i u,
  sub (bullet i TyLeaf) u ->
  exists j,
    i <= j /\ equal (bullet j TyLeaf) u.
Proof.
  intro i.
  myGeneralize (bullet i TyLeaf).
  intros t e u s; induction s.
  assert (i = 0). apply leaf_leaf_inversion. auto. exists 0. split. omega. auto.
  false. eapply bullet_leaf_arrow. eauto. 
  forwards (leaf_bullet_inversion i u e).

 intros [ im1 [ foo bar ]]. auto.
*)

Lemma leaf_inversion:
  forall i1 i2, sub (bullet i1 TyLeaf) (bullet i2 TyLeaf) -> i1 <= i2.
Proof.
  intros i1 i2 h.
  generalize (sub_dsub _ _ h 0); intro hop. repeat (rewrite push_bullet_is_bullet in *). simpl in *.
  myInversions hop.
  myBulletInjective H.
  myBulletInjective H0.
  omega.
Qed.

Lemma arrow_inversion:
  forall i1 u1 v1 i2 u2 v2,
    sub (bullet i1 (TyArrow u1 v1)) (bullet i2 (TyArrow u2 v2)) ->
    exists k,
    sub (bullet i2 u2) (bullet (k + i1) u1) /\ sub (bullet (k + i1) v1) (bullet i2 v2).
Admitted. (* TEMPORARY *)
(*
Proof.
  intros u1 v1 u2 v2 h.
  generalize (sub_dsub _ _ h 0); intro hop. simpl in *.
*)

Lemma bullet_push:
  forall u m i, bullet m TyLeaf = push i u -> m >= i /\ u = bullet (m - i) TyLeaf.
Proof.
  induction u; intros.
  simpl in *. myBulletInjective H. subst. myForwardOmega (i - i = 0).
  myImpossible.
  simpl in *. 
  generalize (IHu _ _ H). intros [ ? ? ]. subst.
  split. omega.
  myForwardOmega (m - i = 1 + (m - (i + 1))).
Qed.
Require Import Coq.Program.Equality.

Lemma bullet_inversion_dsub:
(*
  forall l r, dsub l r -> forall u v i, l = push i u -> r = push i v -> forall j, dsub (push j u) (push j v).
*)
  forall u v i, dsub (push i u) (push i v) -> forall j, dsub (push j u) (push j v).
Proof.
  intros u v i h.
  dependent induction h; intros.
  (* Leaf Leaf *)
  generalize (bullet_push _ _ _ H0).
  generalize (bullet_push _ _ _ H1).
  intros [ ? ? ] [ ? ?]. subst.
  repeat (rewrite push_bullet_is_bullet).
  econstructor. omega.
  (* Arrow Arrow *)
  myPushArrow H.
  myPushArrow H0.
  repeat (rewrite push_bullet_is_push). simpl in *.
  eapply (DSubArrow k); repeat (rewrite push_push in *);
  myForwardOmega (k + (j + x) = j + (k + x));
  repeat (rewrite <- (push_push _ j));
  [ eapply (IHh1 i) | eapply (IHh2 i) ];
  try myForwardOmega (k + (i + x) = i + (k + x));
  eauto using push_push. (* grosse artillerie! *)
Qed.

Lemma bullet_inversion:
  forall u v, sub (TyBullet u) (TyBullet v) -> sub u v.
Proof.
  intros u v h.
  eapply dsub_push_push_sub.
  generalize (sub_dsub _ _ h 0); intro hop. simpl in *.
  eauto using bullet_inversion_dsub.
Qed.

(* Definition of the constraint generator. *)

(* Require Import Coq.Arith.Compare_dec. *)

Require Import Coq.Arith.Wf_nat.

Fixpoint gen i1 t1 i2 t2 : Prop :=
  match t1, t2 with
    TyBullet t1, _      => gen (i1 + 1) t1 i2 t2
  | _, TyBullet t2      => gen i1 t1 (i2 + 1) t2
  | TyLeaf, TyLeaf      => i1 <= i2 (* if le_lt_dec i1 i2 then 0 else i1 - i2 *)
  | TyArrow t11 t12,
    TyArrow t21 t22     => exists k,
                             gen i2 t21 (i1 + k) t11 /\ gen (i1 + k) t12 i2 t22
  | TyLeaf, TyArrow _ _
  | TyArrow _ _, TyLeaf => False
  end.

Theorem gen_correct:
  forall i1 t1 i2 t2,
    gen i1 t1 i2 t2 <-> sub (bullet i1 t1) (bullet i2 t2).